#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <memory.h>

#include "memoryalloc.h"
#include "hashtable.h"
#include "hashes.h"

struct HashtableType
StringTableType = {
  StringHash,
  StringCopy,
  StringCopy,
  AtomDelete,
  AtomDelete,
  StringEqual,
  NULL
};

struct HashtableType
LiteralToStringTableType = {
  LiteralHash,
  LiteralCopy,
  StringCopy,
  LiteralDelete,
  AtomDelete,
  LiteralEqual,
  NULL
};

size_t LiteralHash(void *data)
{
  return (size_t)data;
}

size_t StringHash(void *data)
{
  size_t result = 0;
  char * text = (char *)data;

  if ((text == 0) || (text[0] == 0))
  {
    return 3;
  }

  while (*text)
  {
    result = (11 * result) + (*((unsigned char const *)text));
    ++text;
  }
  return result;
}

void *LiteralCopy(void *data)
{
  return data;
}

void *StringCopy(void *data)
{
  size_t len;
  void *result;

  if (data == NULL)
  {
    data = "";
  }
  len = strlen((char *)data) + 1;
  result = SafeMalloc(len * sizeof (char));
  memcpy(result, data, len * sizeof (char));
  return result;
}

void LiteralDelete(void *data)
{
  (void)data;
  /* do nothing. */
}

void AtomDelete(void *data)
{
  if (data)
  {
    free(data);
  }
}

unsigned LiteralEqual(void *first, void *second)
{
  return first == second;
}

unsigned StringEqual(void *first, void *second)
{
  char *f = (char *)first;
  char *s = (char *)second;
  if (f == s)
  {
    return 1;
  }
  if (f == NULL)
  {
    f = "";
  }
  if (s == NULL)
  {
    s = "";
  }
  return strcmp(f, s) == 0;
}
