#include <stdio.h>
#include <malloc.h>

#include "memoryalloc.h"
#include "hashtable.h"

void HashtableMake(
  struct Hashtable * h,
  struct HashtableType * type
)
{
  size_t i;

  h->data = (struct List **)SafeMalloc(INITIAL_N * sizeof (struct List *));
  for (i = 0; i < INITIAL_N; ++i)
  {
    h->data[i] = NULL;
  }
  h->capacity = INITIAL_N;
  h->size = 0;
  h->type = type;
}

void FreeNode(struct Hashtable *h, struct List *node)
{
  if (node->data != h->type->invalid)
  {
    h->type->DeleteData(node->data);
    h->type->DeleteKey(node->key);
  }
  free(node);
}

void HashtableDestroy(struct Hashtable * h)
{
  size_t i;
  struct List *node;
  struct List *next;

  /*
    //size_t colCount = 0;
    //size_t freeCount = 0;
    //size_t collided = 0;
    //unsigned col;
  */

  for (i = 0; i < h->capacity; ++i)
  {
    /*
      //col = 0;
      //if (h->data[i] == NULL)
      //{
      //  freeCount++;
      //}
      //else if (h->data[i]->next != NULL)
      //{
      //  col = 1;
      //  colCount++;
      //}
    */
    for (node = h->data[i]; node;)
    {
      /*
        //if (col)
        //{
        //  collided++;
        //}
      */
      next = node->next;
      FreeNode(h, node);
      node = next;
    }
  }
  free(h->data);
  /*
    //printf("cap = %u, ave = %f, col = %u, col pts = %u, free = %u\n",
    //  (unsigned)h->capacity,
    //  ((float)collided / (float)colCount),
    //  (unsigned)collided, (unsigned)colCount, (unsigned)freeCount);
  */
}

struct Hashtable * HashtableNew(struct HashtableType * type)
{
  struct Hashtable * result;

  result = (struct Hashtable *)SafeMalloc(sizeof (struct Hashtable));
  HashtableMake(result, type);
  return result;
}

void HashtableDelete(struct Hashtable * h)
{
  HashtableDestroy(h);
  free(h);
}

void * HashtableGet(struct Hashtable * h, void * key)
{
  size_t hashCode;
  struct List *node;

  hashCode = h->type->Hash(key) % h->capacity;
  for (node = h->data[hashCode]; node; node = node->next)
  {
    if (h->type->EqualKeys(node->key, key))
    {
      return node->data;
    }
  }
  return h->type->invalid;
}

void HashtableRehash(struct Hashtable * h)
{
  struct Hashtable oldTable;
  size_t i;
  struct List *node;
  const size_t N = h->capacity;
  const size_t NewN = N + N;


  if (NewN <= N)
  {
    FatalError("Too many entries in hashtable");
  }
  oldTable.data = h->data;
  oldTable.capacity = N;
  oldTable.type = h->type;
  h->data = (struct List **)SafeMalloc(NewN * sizeof (struct List *));
  for (i = 0; i < NewN; ++i)
  {
    h->data[i] = NULL;
  }
  h->capacity = NewN;
  h->size = 0;
  for (i = 0; i < N; ++i)
  {
    for (node = oldTable.data[i]; node; node = node->next)
    {
      HashtableAddInternal(h, node->key, node->data, 0);
      node->data = h->type->invalid;
    }
  }
  HashtableDestroy(&oldTable);
}

unsigned HashtableAddInternal(struct Hashtable * h, void *key, void *data, unsigned copy)
{
  size_t hashCode;
  struct List *node;

  hashCode = h->type->Hash(key) % h->capacity;
  for (node = h->data[hashCode]; node; node = node->next)
  {
    if (h->type->EqualKeys(node->key, key))
    {
      return 0;
    }
  }
  if (h->size >= ((size_t)(0.9 * h->capacity)))
  {
    HashtableRehash(h);
    /*
      //printf("rehash, new capacity is : %u\n", ((unsigned)h->capacity));
    */
  }
  node = (struct List *)SafeMalloc(sizeof (struct List));
  if (copy)
  {
    node->key = h->type->CopyKey(key);
    node->data = h->type->CopyData(data);
  }
  else
  {
    node->key = key;
    node->data = data;
  }
  node->next = h->data[hashCode];
  h->data[hashCode] = node;
  ++h->size;
  return 1;
}

unsigned HashtablePut(struct Hashtable * h, void *key, void *data)
{
  return HashtableAddInternal(h, key, data, 1);
}

unsigned HashtableRemove(struct Hashtable * h, void *key)
{
  size_t hashCode;
  struct List **pp;
  struct List *toDelete;

  hashCode = h->type->Hash(key) % h->capacity;
  pp = &(h->data[hashCode]);
  while (*pp)
  {
    if (h->type->EqualKeys((*pp)->key, key))
    {
      toDelete = *pp;
      *pp = toDelete->next;
      FreeNode(h, toDelete);
      --h->size;
      return 1;
    }
    pp = &((*pp)->next);
  }
  return 0;
}
