#ifndef HASHTABLE_H
#define HASHTABLE_H

#define INITIAL_N (1021)

typedef size_t (*HashType)(void *data);
typedef void * (*CopyKeyType)(void *data);
typedef void * (*CopyDataType)(void *data);
typedef void (*DeleteKeyType)(void *data);
typedef void (*DeleteDataType)(void *data);
typedef unsigned (*EqualKeysType) (void *first, void *second);

struct List
{
  void *key;
  void *data;
  struct List *next;
};

struct HashtableType
{
  HashType Hash;
  CopyKeyType CopyKey;
  CopyDataType CopyData;
  DeleteKeyType DeleteKey;
  DeleteDataType DeleteData;
  EqualKeysType EqualKeys;
  void *invalid;
};

struct Hashtable
{
  struct HashtableType *type;
  struct List **data;
  size_t capacity;
  size_t size;
};

struct Hashtable * HashtableNew(struct HashtableType * type);
void HashtableDelete(struct Hashtable * h);

void HashtableMake(struct Hashtable * h, struct HashtableType * type);
void HashtableDestroy(struct Hashtable * h);

void * HashtableGet(struct Hashtable * h, void * key);
unsigned HashtablePut(struct Hashtable * h, void * key, void * data);
unsigned HashtableRemove(struct Hashtable * h, void * key);

void HashtableRehash(struct Hashtable * h);
unsigned HashtableAddInternal(struct Hashtable * h, void * key, void *data, unsigned copy);

void FreeNode(struct Hashtable *h, struct List *node);

#endif
