#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "memoryalloc.h"
#include "hashtable.h"
#include "hashes.h"
#include "letlittolithash.h"
#include "stringops.h"
#include "state.h"
#include "stateset.h"
#include "ast.h"

void StringTests()
{
  char *a, *b, *c, *d;
  a = "aab";
  b = "aabb";
  c = "aax";
  d = AllocateAnotherString(a);
  GetCommonPrefix(a, b, d);
  printf("first d = %s\n", d);
  GetCommonPrefix(d, c, d);
  printf("second d = %s\n", d);
  free(d);
}

/*

//  char *in[]  = {"adr", "abe", "acf", "bdr", "bbe", "bcf"};
//  char *out[] = {"112", "13",  "13",  "12",  "3"  , "3"};
//  for (i = 0; i < 6; ++i)
//  {
//    ASTAddPair(ast, in[i], out[i]);
//  }
//  ASTRemovePair(ast, in[5]);
//  ASTAddPair(ast, in[5], out[5]);
//  {
//    ASTPrint(ast, stdout);
//  }

*/

/**
 * Make sure this is at least 5, or the strcpy in
 * ReadString() will overflow.
 */
#define BUFFER_LENGTH (1024)

static void ReadString(char * buf)
{
  putc(']', stdout);
  fflush(stdout);
  if (fgets(buf, BUFFER_LENGTH - 1, stdin) == NULL)
  {
    strcpy(buf, "exit");
  }
  buf[BUFFER_LENGTH - 2] = '\0';
  buf[BUFFER_LENGTH - 1] = '\0';
}

static char * SkipWord(char * text)
{
  while (*text)
  {
    if (!(*text >= FIRST_LETTER && *text <= LAST_LETTER))
    {
      break;
    }
    ++text;
  }
  *text = ' ';
  return text;
}

static char * AdvanceWord(char * text)
{
  char *end;

  while (*text)
  {
    if (*text >= FIRST_LETTER && *text <= LAST_LETTER)
    {
      break;
    }
    ++text;
  }
  if (!*text)
  {
    return NULL;
  }
  end = text;
  while (*end >= FIRST_LETTER && *end <= LAST_LETTER)
  {
    ++end;
  }
  *end = '\0';
  return text;
}

int main()
{
  static char FILE_NAME[] = "output.txt";

  static char LINE[] = "%s\n";

  static char CMD_PUT[] = "put";
  static char CMD_GET[] = "get";
  static char CMD_REMOVE[] = "remove";
  static char CMD_PRINT[] = "print";
  static char CMD_SAVE[] = "save";
  static char CMD_HELP[] = "help";
  static char CMD_EXIT[] = "exit";

  static char MESSAGE_PUT[] = "put key value - Puts a key/value pair to the AST";
  static char MESSAGE_GET[] = "get key - Gets a value from the AST by the given key";
  static char MESSAGE_REMOVE[] = "remove key - Removes a value from the AST by the given key";
  static char MESSAGE_PRINT[] = "print - Prints the AST";
  static char MESSAGE_SAVE[] = "save - Saves the AST to output.txt";
  static char MESSAGE_HELP[] = "help - Prints this information";
  static char MESSAGE_EXIT[] = "exit - Exits the program";

  static char MESSAGE_PUT_ERROR[] = "The strings are too large";
  static char MESSAGE_PUT_ERROR2[] = "The key is already present";
  static char MESSAGE_GET_ERROR[] = "This key was not found";
  static char MESSAGE_SAVE_ERROR[] = "Cannot save output.txt";

  static char MESSAGE_ERROR[] = "Enter 'help' for information";

  struct AST *ast;
  struct State *state;
  char buf[BUFFER_LENGTH];
  char putGetBuf[MEMORY_BLOCK];
  char *next;
  char *first;
  char *second;
  FILE *in;
  size_t i;
  size_t length;

  StateInitTables();

  ast = ASTNew(NULL, NULL, 0);

  for (;;)
  {
    ReadString(buf);
    next = AdvanceWord(buf);
    if (next)
    {
      if (strcmp(next, CMD_PUT) == 0)
      {
        first = next;
        next = SkipWord(next);
        next = AdvanceWord(next);
        if (!next)
        {
          printf(LINE, MESSAGE_PUT);
        }
        else if (ASTIsKeyPresent(ast, next))
        {
          printf(LINE, MESSAGE_PUT_ERROR2);
        }
        else
        {
          strcpy(putGetBuf, next);
          first = putGetBuf;
          next = SkipWord(next);
          second = AdvanceWord(next);
          if (second == NULL)
          {
            second = empty;
          }
          if (strlen(first) > MAX_STRING || strlen(second) > MAX_STRING)
          {
            printf(LINE, MESSAGE_PUT_ERROR);
          }
          else
          {
            ASTAddPair(ast, first, second);
          }
        }
      }
      else if (strcmp(next, CMD_GET) == 0)
      {
        first = next;
        next = SkipWord(next);
        next = AdvanceWord(next);
        if (!next)
        {
          printf(LINE, MESSAGE_GET);
        }
        else
        {
          putGetBuf[0] = '\0';
          state = ast->initialState;
          length = strlen(next);
          for (i = 0; i < length; ++i)
          {
            if (state == NULL)
            {
              break;
            }
            strcat(putGetBuf, StateGetOutput(state, next[i]));
            state = StateGetTransition(state, next[i]);
          }
          if (i < length || state == NULL || (!StateIsFinal(state)))
          {
            printf(LINE, MESSAGE_GET_ERROR);
          }
          else
          {
            strcat(putGetBuf, StateGetStateOutput(state));
            printf(LINE, putGetBuf);
          }
        }
      }
      else if (strcmp(next, CMD_REMOVE) == 0)
      {
        first = next;
        next = SkipWord(next);
        next = AdvanceWord(next);
        if (!next)
        {
          printf(LINE, MESSAGE_REMOVE);
        }
        else if (!ASTIsKeyPresent(ast, next))
        {
          printf(LINE, MESSAGE_GET_ERROR);
        }
        else
        {
          ASTRemovePair(ast, next);
        }
      }
      else if (strcmp(next, CMD_PRINT) == 0)
      {
        ASTPrint(ast, stdout, 0);
      }
      else if (strcmp(next, CMD_SAVE) == 0)
      {
        in = fopen(FILE_NAME, "w");
        if (!in)
        {
          printf(LINE, MESSAGE_SAVE_ERROR);
        }
        else
        {
          ASTPrint(ast, in, 0);
          fclose(in);
        }
      }
      else if (strcmp(next, CMD_HELP) == 0)
      {
        printf(LINE, MESSAGE_PUT);
        printf(LINE, MESSAGE_GET);
        printf(LINE, MESSAGE_REMOVE);
        printf(LINE, MESSAGE_PRINT);
        printf(LINE, MESSAGE_SAVE);
        printf(LINE, MESSAGE_HELP);
        printf(LINE, MESSAGE_EXIT);
      }
      else if (strcmp(next, CMD_EXIT) == 0)
      {
        break;
      }
      else
      {
        printf(LINE, MESSAGE_ERROR);
      }
    }
  }

  ASTDelete(ast);

  StateDestroyTables();

  printf("Press enter to exit...");
  fgets(buf, MAX_STRING, stdin);

  return 0;
}
