#ifndef STATE_H
#define STATE_H

/*
   Only values between these
   two letters, inclusive,
   are permitted
*/

/*
#define FIRST_LETTER ((size_t)('a'))
#define LAST_LETTER ((size_t)('z'))
*/

#define FIRST_LETTER ((size_t)('#'))
#define LAST_LETTER ((size_t)('~'))

#define LETTERS_COUNT ((size_t)(LAST_LETTER - FIRST_LETTER + 1))
#define MAX_STRING ((size_t)100)
#define MEMORY_BLOCK ((size_t)(MAX_STRING + 1))

#define CHAR_VAR_TO_INDEX(X) (((size_t)(*((unsigned char *)(&(X))))) - FIRST_LETTER)

extern struct HashtableType StateToStateTableType;

struct State;

struct TempStateData
{
  char psy[MEMORY_BLOCK];
  char lambda[LETTERS_COUNT][MEMORY_BLOCK];
  struct State *delta[LETTERS_COUNT];
};

struct PermanentStateData
{
  char *psy;
};

union StateData
{
  struct TempStateData *tempStateData;
  struct PermanentStateData permanentStateData;
};

struct State
{
  union StateData stateData;
  size_t backLinks;
  size_t hashCode;
  int id;
  int copyOf;
  unsigned char hashCalculated;
  unsigned char isFinal;
  unsigned char isTemp;
  unsigned char forceModifyChildrenOnDestroy;
};

void StateInitTables();
void StateDestroyTables();

struct State *StateNew(unsigned isTemp);
void StateDelete(struct State *state);
void StateTerminate(struct State *state);

void StateMake(struct State *state, unsigned isTemp);
void StateDestroy(struct State *state, unsigned modifyChildren);

unsigned StateIsFinal(struct State *state);
void StateSetFinal(struct State *state, unsigned isFinal);
struct State *StateGetTransition(struct State *state, char letter);
void StateSetTransition(struct State *state, char letter, struct State *to);
int StateGetID(struct State *state);
char *StateGetStateOutput(struct State *state);
void StateSetStateOutput(struct State *state, char *psy);
char *StateGetOutput(struct State *state, char letter);
void StateSetOutput(struct State *state, char letter, char *lambda);
unsigned StateHasBackLinks(struct State *state);
void StateSetClear(struct State *state);
void StateCopyInto(struct State *state, struct State *other);
struct State *StateGetCopy(struct State *state, unsigned isTemp);
void StateGetCommonPrefix(struct State *state, char prefix[MEMORY_BLOCK]);
void StateRemoveCommonPrefix(struct State *state, size_t size);
size_t StateHashCode(struct State *state);
unsigned StateEquals(struct State *state, struct State *other);
void StatePrint(struct State *state, FILE *file);

#endif
