#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "memoryalloc.h"
#include "hashtable.h"
#include "hashes.h"
#include "letlittolithash.h"
#include "stringops.h"
#include "vector.h"
#include "state.h"
#include "stateset.h"

void StateSetMake(struct StateSet *set, struct State *tempStates)
{
  HashtableMake(&set->states, &StateToStateTableType);
  set->tempStates = tempStates;
}

void StateSetDestroy(struct StateSet *set)
{
  if (set)
  {
    /*
     * This deletes all permanent states
     */
    HashtableDestroy(&set->states);
  }
}

struct StateSet *StateSetNew(struct State *tempStates)
{
  struct StateSet *set = (struct StateSet *)
    SafeMalloc(sizeof (struct StateSet));
  StateSetMake(set, tempStates);
  return set;
}

void StateSetDelete(struct StateSet *set)
{
  if (set)
  {
    StateSetDestroy(set);
    free(set);
  }
}

struct State *StateSetGetMember(struct StateSet *set, struct State *state)
{
  return (struct State *)HashtableGet(&set->states, state);
}

void StateSetInsert(struct StateSet *set, struct State *state)
{
  HashtablePut(&set->states, state, state);
}

void StateSetRemove(struct StateSet *set, struct State *state)
{
  /*
   * This deletes the equivalent of the state --
   * this may be the state itself!
   */
  HashtableRemove(&set->states, state);
}

struct State *StateSetFindMinimized(struct StateSet *set, struct State * s)
{
  struct State *r;

  r = StateSetGetMember(set, s);
  if (r != s)
  {
    if (r == NULL)
    {
      r = StateGetCopy(s, 0);
      StateSetInsert(set, r);
    }
    StateSetClear(s);
  }

  return r;
}

void StateSetDePropagate(struct StateSet *set, char *currentWord, size_t lastIndex)
{
  char prefix[MEMORY_BLOCK];
  size_t length;
  size_t i;

  for (i = lastIndex; i > 0; --i)
  {
    StateGetCommonPrefix(&set->tempStates[i], prefix);
    if (*prefix)
    {
      length = strlen(prefix);
      Concatenate(
        StateGetOutput(&set->tempStates[i - 1], currentWord[i - 1]),
        prefix,
        prefix);
      StateSetOutput(&set->tempStates[i - 1], currentWord[i - 1], prefix);
      StateRemoveCommonPrefix(&set->tempStates[i], length);
    }
  }
}

struct State *StateSetClone(struct StateSet *set, struct State *parent, char c, struct State *from, struct State *to)
{
  if (parent != NULL)
  {
    StateSetTransition(parent, c, to);
  }

  StateCopyInto(from, to);

  if (!StateHasBackLinks(from))
  {
    /*
     * This deletes from
     */
    from->forceModifyChildrenOnDestroy = (unsigned char)1;
    StateSetRemove(set, from);
  }

  return to;
}
