#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "memoryalloc.h"
#include "Vector.h"

void VectorMake(struct Vector * v)
{
  if (v)
  {
    v->capacity = 0;
    v->length = 0;
    v->data = NULL;
  }
}

void VectorDestroy(struct Vector * v)
{
  if (v)
  {
    if (v->data)
    {
      free(v->data);
    }
    /* Prepare for future use : */
    VectorMake(v);
  }
}

static void VectorSimpleAdd(struct Vector * v, int newData)
{
  v->data[v->length] = newData;
  v->length++;
}

static int * VectorResizeArray(
  int * oldData,
  size_t oldSize,
  size_t newSize
)
{
  int * newData = (int *)malloc(newSize * sizeof (int));
  if (!newData)
  {
    FatalError(
      "Cannot allocate memory for a Vector : %d\n",
      newSize
    );
  }
  if (oldData)
  {
    memcpy(newData, oldData, oldSize * sizeof (int));
    free(oldData);
  }
  return newData;
}

void VectorAdd(struct Vector * v, int newData)
{
  if (!v)
  {
    FatalError("Trying to add to a NULL Vector\n");
  }
  if (v->capacity > v->length)
  {
    VectorSimpleAdd(v, newData);
  }
  else
  {
    /* Increase the capacity : */
    size_t newSize = v->capacity * 2;
    if (newSize <= v->capacity)
    {
      newSize = v->capacity + 256;
      if (newSize <= v->capacity)
      {
        FatalError(
          "Too many entries in Vector : %d\n",
          v->length
        );
      }
    }
    v->capacity = newSize;
    v->data = VectorResizeArray(v->data, v->length, newSize);
    VectorSimpleAdd(v, newData);
  }
}

size_t VectorIndexOf(struct Vector * v, int data)
{
  size_t i;

  for (i = 0; i < v->length; ++i)
  {
    if (v->data[i] == data)
    {
      return i;
    }
  }
  return NOT_IN_VECTOR;
}

unsigned VectorRemove(struct Vector * v, int oldData)
{
  size_t i;

  if (!v)
  {
    FatalError("Trying to remove from a NULL Vector\n");
  }
  for (i = 0; i < v->length; ++i)
  {
    if (v->data[i] == oldData)
    {
      memmove(v->data + i, v->data + (i + 1), (v->length - (i + 1)) * sizeof (int));
      return 1;
    }
  }
  return 0;
}

void VectorIncrementCapacity(struct Vector * v, size_t size)
{
  size_t newSize;
  if (!v)
  {
    FatalError("Trying to add to a NULL Vector\n");
  }
  newSize = v->capacity + size;
  if (newSize <= v->capacity)
  {
    FatalError(
      "Too many entries in Vector : %d",
      v->length
    );
  }
  v->capacity = newSize;
  v->data = VectorResizeArray(v->data, v->length, newSize);
}
